# 多线程

多线程的目的

+ 减少CPU的空闲
+ 实现异步任务

Thread类
	t1.join()
		当前线程挂起，等待t1线程执行完再执行
	Thread.yield()
		放弃cpu使用权，建议jvm让步给同级或高优先级的线程先运行。注意：因为是抢占式调度，有可能下次本线程又被JVM调度

**CPU密集型任务**

- 如分词，加密，递归，正则，数学计算
- 最佳线程数 = cpu核数或稍小于cpu核数（为什么小于cpu核数，因为一个java进程启动就至少有2个线程，main和GC线程）

**IO密集型**

- 如网络I/O，文件I/O，socket
- 最佳线程数 = 单个线程空闲时间/繁忙时间 * cpu核数

<span style="color:red;">不适合多线程的场景</span>：如只是单纯的读文件，一个线程足够了，多个线程只会在磁盘io上阻塞

## 线程的2种实现

- class A extends Thread（Thread实现了Runnable接口）
- Class B implements Runnable

## JVM内存模型

![image-20200113104540948](..\assets\image-20200113104540948.png)

1. 多个线程共享的变量，存储在主内存
2. 线程自己的工作内存中保存了主内存共享变量的副本，线程要操作这些共享变量，只能通过操作工作内存中的副本来实现，操作完再同步回主内存
3. 如果线程a读取一个变量到工作内存，修改了变量，此时还没有同步回主内存，此时变量对于线程b来说，是不可见的。如何让线程b立即可见呢？变量用volatile修饰

## volatile

```java
volatile int i = "123";
```

- 每次修改volatile变量都会同步到主存中
- 每次读取volatile变量都强制从主存读取最新的值
- volatile只保证变量的**内存可见性**，**不能保证**在此变量上操作的**原子性**，所以==不能起到互斥==的作用（*volatile是轻量级的同步机制*）i++依然线程不安全

## 线程封闭

就是使用局部变量，它们存在于**线程栈**中，其他线程是无法访问的，这样就安全了。比如ThreadLocal

## <span style="color:#fff; background:#009688;">ThreadLocal</span>

ThreadLocal，顾名思义，其存储的变量是线程私有的，为每一个使用该变量的线程都提供一个==副本==，以避免线程争用或者同步问题，例如使用ThreadLocal来为每个线程维持一个redis连接（生产中这也许不是一个好的方式，还是推荐专业的连接池）

例如，SimpleDateFormat不是线程安全的，那么我们通常用两种做法：

- 每次用到的时候new一个SimpleDateFormat对象，使用完丢弃，交给gc
- 每个线程维护一个SimpleDateFormat实例，线程运行期间不重复创建

那么无论从执行效率还是内存占用方面，我们都倾向于使用后者，即线程私有一个SimpleDateFormat对象，这时候，ThreadLocal就是很好的应用，示例代码如下：

```java
public class TestTask implements Runnable {
  private static ThreadLocal<SimpleDateFormat> holder = new ThreadLocal<SimpleDateFormat>() {
    @Override
    protected SimpleDateFormat initialValue() {
      return new SimpleDateFormat("yyyyMMdd");
    }
  };

  @Override
  public void run() {
    //每个线程只有一个SimpleDateFormat对象
    SimpleDateFormat formatter = holder.get();
    //holder.remove();
  }
}
```

### 源码

```java
public void set(T value) {
	Thread t = Thread.currentThread();
	ThreadLocalMap map = getMap(t);
	if (map != null)
		map.set(this， value);
	else
		createMap(t， value);
}
ThreadLocalMap getMap(Thread t) {
	return t.threadLocals;
}

static class ThreadLocalMap {
  ThreadLocalMap(ThreadLocal firstKey, Object firstValue) {
    table = new Entry[INITIAL_CAPACITY];
    int i = firstKey.threadLocalHashCode & (INITIAL_CAPACITY - 1);
    table[i] = new Entry(firstKey, firstValue);
    size = 1;
    setThreshold(INITIAL_CAPACITY);
  }
  static class Entry extends WeakReference<ThreadLocal> {
    /** The value associated with this ThreadLocal. */
    Object value;

    Entry(ThreadLocal k, Object v) {
      super(k);
      value = v;
    }
  }
}
```

### InheritableThreadLocal

在主线程中创建InheritableThreadLocal实例后，通过new Thread()方式创建子线程后，子线程就可以获取到主线程设置的值了

### 内存泄露

![img](D:\Program Files\wanglei_notebook\assets\threadlocal.png)

每一个Thread维护一个ThreadLocalMap，key是ThreadLocal实例（ThreadLocal的弱引用），value是具体存储的Object。如果ThreadLocal没有外部强引用，当发生垃圾回收时，这个ThreadLocal一定会被回收(弱引用的特点是不管当前内存空间足够与否，GC时都会被回收)，这样就会导致ThreadLocalMap中出现key为null的Entry，外部将不能获取这些key为null的Entry，如果当前线程一直存活，那么就会存在一条强引用链：Thread Ref -> Thread -> ThreaLocalMap -> Entry -> value，导致value对应的Object一直无法被回收，产生内存泄露。

ThreadLocal的get、set和remove方法都实现了对所有key为null的value的清除，但仍可能会发生内存泄露，因为可能使用了ThreadLocal的get或set方法后发生GC，此后不调用get、set或remove方法，为null的value就不会被清除。

**解决方法**：

1. 每次使用完ThreadLocal都调用它的remove()清除数据
2. 按照JDK建议将ThreadLocal变量定义成private static，这样就一直存在ThreadLocal的强引用，也就能保证任何时候都能通过ThreadLocal的弱引用访问到Entry的value值，进而清除掉

### 隐式传参

ThreadLocal里的变量可以在整个线程存活的过程中随时取用，可以简化编程，不用在每个方法入参都写上了

## 中断

> 如何停止运行的线程？用信号量或中断

```java
public class Thread implements Runnable {
    public void interrupt() {
        //调用t.interrput()的时候，线程t的中断状态会被置位
        //通过Thread.currentThread().isInterrupted() 来检查这个中断状态
    }
    
    private native boolean isInterrupted(boolean ClearInterrupted);
    //将中断标标位clear掉
    public static boolean interrupted() {
        return currentThread().isInterrupted(true);
    }
}
```

中断一个线程只是为了引起该线程的注意，被中断线程可以自行决定如何响应

**线程在I/O操作时被阻塞，可以被中断**

当线程被阻塞时，如Object.wait/condition.await， Thread.join，Thread.sleep以及可中断的I/O方法。调用interrput()，可想而知，没有占用CPU的线程是不可能给自己的中断状态置位的。这就会产生一个InterruptedException，JVM在抛出InterruptedException前会将线程的中断标别位清除（设false）。**抛出异常是为了线程从阻塞状态醒过来**

实现InterruptibleChannel接口的通道是可中断的

- serverSocketChannel.accept
- socketChannel.connect
- socketChannel.open
- socketChannel.read
- socketChannel.write
- fileChannel.read
- fileChannel.write

另一个线程调用该阻塞线程的 interrupt 方法，这将导致该通道被关闭，并且已阻塞线程接将会收到ClosedByInterruptException，并且设置已阻塞线程的中断状态

## 处理非check异常

```java
/**
* Dispatch an uncaught exception to the handler. This method is 
* intended to be called only by the JVM.
*/
private void dispatchUncaughtException(Throwable e) {
    getUncaughtExceptionHandler().uncaughtException(this， e);
}

/**
	JVM寻找handler的顺序：
  当前线程对象的uncaughtExceptionHandler -> ThreadGroup中的uncaughtExceptionHandler -> default uncaughtExceptionHandler -> console output
*/
public UncaughtExceptionHandler getUncaughtExceptionHandler() { 
	return uncaughtExceptionHandler != null ? uncaughtExceptionHandler : group;
}

/**
 *group就是ThreadGroup，看看他的uncaughtException方法
 */
public void uncaughtException(Thread t， Throwable e) {
	if (parent != null) {
	    parent.uncaughtException(t， e);
	} else {
		Thread.UncaughtExceptionHandler ueh = Thread.getDefaultUncaughtExceptionHandler();
          if (ueh != null) {
          	ueh.uncaughtException(t， e);
          } else if (!(e instanceof ThreadDeath)) {
			System.err.print("Exception in thread \"" + t.getName() + "\" ");
          	e.printStackTrace(System.err);
          }
     }
}

//使用
thread.setUncaughtExceptionHandler(new MyHandler());
thread.setDefaultUncaughtExceptionHandler(new MyHandler());
```

## Fail-Fast

速错机制，当使用Iterator迭代过程中，有其他线程或本线程修改了集合内容，将导致Iterator抛出ConcurrentModificationException，来阻止错误继续发生。修改动作不包括Iterator本身调用remove()

```java
//ConcurrentModificationException不一定要并发下才会产生
for (Iterator it = map.entrySet().iterator(); it.hasNext();) {
    Map.Entry entry = (Map.Entry) it.next();
    map.remove(entry.getKey()); // it.remove()不会抛出异常
}
```

**原理**：通过modCount记录集合内容的修改次数，每次集合的内容的修改都会增加这个变量。迭代器初始化过程中会将这个值赋给迭代器的expectedModCount，见AbstractList.java源码

```java
    /**
     修改指的是能够改变集合size的操作，如果modCount改变了，则会在next、previous、remove、set、add方法调用的时候抛出异常
     * This provides fail-fast behavior， rather than non-deterministic behavior in
     * the face of concurrent modification during iteration.
     */
    protected transient int modCount = 0;
```

## 锁

### <span style="color:#fff; background:red;">同步锁synchronized</span>

synchronized是JVM层面的机制，java方法或同步块编译后首尾有添加两条指令：monitorenter和monitorexit（每个对象都有一个监视器锁,这个监视器锁同一时刻只能被一个线程获取，Monitor Lock可重进入，线程在试图获取它自己占有的monitor lock，直接获取 ）

当线程执行到一个对象的任意一个synchronized代码之前，会自动获得该对象的监视器锁

synchronized会把一个对象的所有同步方法和同步块看成一个整体，一旦有一个方法或块加锁了，就会锁住整体。好像效率低了点，因为可能它们之间没关联

**线程释放锁的时机**

1. synchronized方法或代码块正常退出后，**自动**释放
2. throw exception，**自动**释放



- 静态同步块属于类，对类加锁，会限制这个类实例化的所有对象的访问
- 非静态同步快属于对象实例，不会影响类其他对象的访问

```java
synchronized(Hashtable.class){ //锁的是类
}
```

**阻塞线程**

```java
private final String lock = "LOCK";//对象
  
public void produce(){
    synchronized (lock){ //拿到对象锁，才能往下执行
        while(queue.full()){//队列慢了
            try {
                lock.wait();// 当前线程阻塞
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            //当其他线程执行了lock.notify()方法后，唤醒等待的线程，才能执行到这行
        }
    }
}
```

### <span style="color:#fff; background:red;">重入锁ReentrantLock</span>

性能和synchronized相当

**线程释放锁的时机**

1. 需要手动释放锁，若没有手动释放锁，可能导致死锁。即使异常也不释放锁的哦~:smiley:
2. 在finally调用unlock

```java
interface Lock {
	void lock(); //不响应中断,一旦死锁，会陷入永久等待
	void lockInterruptibly() throws InterruptedException;//响应中断
	void unlock();
	
	// 如果当前锁未被其他线程占用，则获取锁，并返回true；否则返回false。方法立即返回
	boolean tryLock();
    // 等待一段时间如果还不能获取锁就返回，不会造成死锁，synchronized就没有这个强大的功能
	boolean tryLock(long time，TimeUnit unit);
	Condition newCondition() throws UnsupportedOperationException;
}
```

必须在同步块中执行，否则会报错

**:question:一个lock有多个condition有啥好处**

+ **可以精确唤醒，而不像synchronized那样随便唤醒一个线程或者全部线程**
+ wait()只有一个线程阻塞队列，notifyAll会唤起这个阻塞队列下的所有线程，
+ 而使用condition，可以实现多个线程阻塞队列（每个condition有一个线程阻塞队列），signalAll只会唤起某个阻塞队列下的线程

| Condition   | Object实例  | 作用                                                         |
| ----------- | ----------- | ------------------------------------------------------------ |
| await()     | wait()      | 当线程执行这个方法后，JVM会将此线程放入和该锁对象关联的wait sets中，并**释放锁**。**前提**是线程必须先获得锁，所以必须在同步块中执行 |
| signal()    | notify()    | 其他线程执行这个方法时，JVM会从wait sets中随机唤醒一个线程，移到锁池。该线程可以在锁池**再次竞争**对象的monitor。抢到锁后就回到上次的代码位置继续执行 |
| signalAll() | notifyAll() | 唤醒所有的等待Thread                                         |

```java
// 然后再看这把锁上有哪些条件 
static Lock lock = new ReentrantLock();
static Condition full = lock.newCondition(); //producer的条件
static Condition empty = lock.newCondition();//consumer条件

class Producer implements Runnable{
    public void run(){
        while(true){
            if(lock.tryLock()){ // 争锁
                if(队列满){
                  full.await(); // 阻塞等待：设置队列满的条件（就是个变量），线程执行到这行代码，会释放锁进入等待池 
                }else{
                    //生产消息
                    empty.signalAll();// 谁在empty.await()上阻塞等待，就精确唤醒这些线程
                }
                
            }
            // sleep 避免一直占用CPU
        }
    }
}
Consumer implements Runnable{
    public void run(){
        while(true){
            if(lock.tryLock()){ // 争锁
                if(队列空) {
                  empty.await(); // 阻塞等待
                }else{
                    //消费消息
                    full.signalAll();// 谁在full.await()上阻塞等待，就精确唤醒这些线程
                }
                
            }
            // sleep 避免一直占用CPU
        }
    }
}
```

**唤醒线程的策略**

```java
//fair=true  公平锁：  按照申请锁的先后顺序来获得锁
//fair=false 非公平锁：会随机选中一个等待线程（synchronized），会产生线程饥饿现象，但是性能高
new RenentrantLock(boolean fair)
```

### 读写锁ReadWriteLock

读读不互斥，读写互斥，写写互斥

```java
static ReentrantReadWriteLock lock = new ReentrantReadWriteLock();
static Lock readLock = lock.readLock();
static Lock writeLock = lock.writeLock();
```

### 死锁

发生死锁有什么现象？ 此时CPU占用为0%

**如何避免死锁**？

+ 各线程按照**相同的顺序**对资源加锁

+ 设置timeout，超时了就自动释放锁，如lock.tryLock(3000)

线上分析步骤

1. 用jconsole或kill -3 pid 导出thread dump文件
2. 分析dump文件

~~~java
final Object a = new Object();
final Object b = new Object();
Thread thread1 = new Thread() {
    public void run() {
         synchronized (a) {
            synchronized (b) {// CPU阻塞在这，不会抛异常
                // Do something
            }
        }
    }
};
Thread thread2 = new Thread() {
    public void run() {
        synchronized (b) {
            synchronized (a) {// CPU阻塞在这，不会抛异常
                // Do something
            }
        }
    }
};
~~~

## JUC

concurrency包为什么支持高并发？

concurrent包是用ReentrantLock构建的，ReentrantLock是基于原子变量（如AtomicInteger）构建的，而原子变量是基于CPU支持的CAS无锁操作

### Queue

接口提供的方法

| 获取并删除队首元素 |                            |
| ------------------ | -------------------------- |
| take               | 若队列空，**阻塞执行线程** |
| poll               | 若队列空，返回null         |
| remove             | 若队列空，抛异常           |

| 单纯获取队首元素 |                    |
| ---------------- | ------------------ |
| peek             | 若队列空，返回null |
| element          | 若队列空，抛异常   |

| 添加元素 |                               |
| -------- | ----------------------------- |
| add      | 若队列满，抛异常              |
| offer    | 若队列满，返回false，否则true |
| put      | 若队列满，**阻塞执行线程**    |

**如何设计BlockingQueue的shutdown操作**

只要queue不满，就一直能put元素，如何shutdown一个queue呢？

常用的策略是：对于生产者，插入特殊的end-of-stream或poison对象，并根据使用者获取这些对象的时间来对它们进行解释

#### ArrayBlockingQueue

> 是一个bounded，**blocking queue**。FIFO，主要用于生产者-使用者模式

```java
private final E[] items;    // The queued items  核心数组
private int takeIndex;      // 队列头
private int putIndex;       // 队列尾
private int count;          // Number of items in the queue
private final ReentrantLock lock;
private final Condition notEmpty = lock.newCondition();// 队列非空的等待条件
private final Condition notFull = lock.newCondition(); // 队列满的等待条件
```

- 入队时，items[putIndex] = 元素，putIndex++，count++
- 出队时，返回items[takeIndex]，takeIndex++，count--
- 出队入队时，由ReentrantLock保证多线程访问的一致性

```java
public void put(E e) throws InterruptedException {
    if (e == null) throw new NullPointerException();
    final E[] items = this.items;
    final ReentrantLock lock = this.lock;
    lock.lockInterruptibly();
    try {
        try {
            while (count == items.length)
                notFull.await();   // 队列已满
        } catch (InterruptedException ie) {
            notFull.signal(); // propagate to non-interrupted thread
            throw ie;
        }
        insert(e);
    } finally {
        lock.unlock();
    }
}
```

#### LinkedBlockingQueue

> 基于链表，线程安全的**blocking queue**。FIFO，吞吐量通常高于ArrayBlockingQueue

静态工厂方法Executors.newFixedThreadPool()使用了这个队列

```java
//注意：这里的count是原子类型，与ArrayBlockingQueue中的不同，这是因为LinkedBlockingQueue使用读和写两把锁来控制并发操作，读和写可能同时修改count，而ArrayBlockingQueue只有一把锁用于控制读写操作，所以count对象是普通的
private final AtomicInteger count = new AtomicInteger(0);

//两把锁分别管理head节点和last节点的操作
private final ReentrantLock takeLock = new ReentrantLock();/* 读锁 */
private final Condition notEmpty = takeLock.newCondition();/* 读锁对应的条件 */
private final ReentrantLock putLock = new ReentrantLock(); /* 写锁 */
private final Condition notFull = putLock.newCondition();  /* 写锁对应的条件 */
```

jdk中解释说，这两把锁的控制，是[two lock queue](http://www.cs.rochester.edu/research/synchronization/pseudocode/queues.html)算法的一种实现

LinkedBlockingQueue的迭代器，是线程安全的，在获取元素之前，会对上述读锁和写锁同时加锁，同时，为了防止死锁，读锁和写锁的加解锁顺序，也是经过设计的，代码如下：

```java
void fullyLock() {
    putLock.lock();      // 先加写锁
    takeLock.lock();     // 再加读锁
}

void fullyUnlock() {
    takeLock.unlock();  // 先解锁读锁
    putLock.unlock();   // 再解锁写锁
}
```

**读写操作分离**

LinkedBlockingQueue将读和写分离，可以让读写操作在不干扰对方的情况下，完成各自的功能，提高并发。

#### SynchronousQueue

> 一个不存储元素的**blocking queue**。生产这的插入操作必须等待消费者调用移除操作，否则插入操作一直处于阻塞状态，反之亦然。吞吐量通常高于LinkedBlockingQueue

SynchronousQueue采用了高性能的[Dual stack and Dual queue](http://www.cs.rochester.edu/research/synchronization/pseudocode/duals.html)算法。

静态工厂方法Executors.newCachedThreadPool使用了这个队列，这个线程池根据需要（新任务到来时）创建新的线程，如果有空闲线程则会重复使用，线程空闲了60秒后会被回收。

#### ConcurrentLinkedQueue

> An unbounded thread-safe **non-blocking** queue based on linked nodes. FIFO

```java
volatile E item;
volatile Node<E> next;
Node(E item) {
    //CAS算法,一般情况下，我们是拿不到Unsafe实例的，当然jdk库里面是可以使用的
	UNSAFE.putObject(this, itemOffset, item);
}
boolean casItem(E cmp, E val) {
	return UNSAFE.compareAndSwapObject(this, itemOffset, cmp, val);
}
```

### ConcurrentHashMap

> 线程安全的HashMap，允许并发修改操作。

关键技术：锁分离

1. 将ConcurrentHashMap分成多个Segment段
2. 每个segment就是一个小的hash table，它们有自己的锁。只要多个修改发生在不同的segment上，就可以并发进行
3. 采用二次hash的方式，第一次hash将key映射到对应的segment，第二次hash映射到segment的不同桶中（这样对于map的修改不会锁住整个容器，提高并发。但是二次hash带来的问题是整个hash的过程比HashMap单次hash长，**所以如果不是并发不要使用ConcurrentHashMap**）

有些方法需要跨段，比如size()和containsValue()，它们可能需要锁定整个表而不仅仅是某个segment，这需要按顺序锁定所有segment，操作完毕后，按顺序释放所有segment的锁。这里“按顺序”是很重要的，否则极有可能出现死锁，在ConcurrentHashMap内部，段数组是final的，并且其成员变量实际上也是final的，但是，仅仅是将数组声明为final的并不保证数组成员也是final的，这需要实现上的保证。这可以确保不会出现死锁，因为获得锁的顺序是固定的。不变性是多线程编程占有很重要的地位。

### CopyOnWriteArrayList

`onWrite`：写入的时候

适合**读多写少**，在并发情况下，可以提高并发读性能

ArrayList线程不安全的，CopyOnWriteArrayList是线程安全的，读操作无须加锁

在并发环境下读性能>ArrayList，但插入、删除比ArrayList慢

**Copy-On-Write**（COW|**写时复制**），是一种优化策略。当我们往一个容器添加元素的时候，不直接往当前容器添加，而是先将当前容器进行Copy，复制出一个新的容器，然后往新的容器里添加元素，添加完元素之后，再将原容器的引用指向新的容器。这样可以对容器进行并发的读，而不需要加锁，因为当前容器不会添加任何元素。所以CopyOnWrite容器也是一种读写分离的思想。缺点：内存大

优化
使用批量添加。因为每次添加，容器每次都会进行复制，所以减少添加次数，可以减少容器的复制次数

使用场景

1、百度云 
收藏别人的资源时,并不是真的将资源copy到自己的空间,而是在设置一个link到源文件,并把源文件的引用数+1
对源文件读的时候,直接访问即可
对源文件有write的时候,才copy一份到新的地址上,link重新指向到新的地址
2、github
3、快照技术，如redis的RDB机制，mapreduce的shuffle阶段，spill线程
4、白名单|黑名单

### 线程执行框架

#### Executors

就是一个工具类，能够创建一些线程池，不建议使用，因为它屏蔽了创建的参数

```java
public static Callable<Object> callable(Runnable task) //将Runnable任务转成Callable任务
ExecutorService newSingleThreadExecutor()
ExecutorService newCachedThreadPool()
ExecutorService newFixedThreadPool(int poolSize)
ScheduledExecutorService newSingleThreadScheduledExecutor()
ScheduledExecutorService newScheduledThreadPool(int poolSize)
```

#### ExecutorService

```java
public interface ExecutorService extends Executor {
	void shutdown();
	List<Runnable> shutdownNow();
	boolean isShutdown();
	boolean isTerminated();// 是否处于TERMINATED状态
	boolean awaitTermination(long timeout， TimeUnit unit) throws InterruptedException;//block until all the task finished
	<T> Future<T> submit(Callable<T> task);
	<T> Future<T> submit(Runnable task， T result);
	Future<?> submit(Runnable task);
	<T> List<Future<T>> invokeAll(Collection<? extends Callable<T>> tasks)
}
```

#### CompletionService

参考https://my.oschina.net/u/4149877/blog/4484414

CompletionService与ExecutorService都可以用来执行线程池的任务，不同点就是二者判断任务完成的效果。

由于Future 结果没有完成，调用 get() 方法，程序会阻塞在那里，直至获取返回结果

用completion.take()，就是哪个任务先完成，就会获取到相应的任务

```java
ExecutorService service = Executors.newCachedThreadPool();
CompletionService<String> completion=new ExecutorCompletionService<String>(service);
List<Future> futures = new ArrayList<Future<Integer>>();
futures.add(completion.submit(A));
futures.add(completion.submit(B));
futures.add(completion.submit(C));
futures.add(completion.submit(D));
for (int i=0; i<futures.size(); i++) {
    Integer result = completion.take().get();
    // 其他业务逻辑
}
service.shutdown();
```

![img](..\assets\e1e33229-d043-49b2-8963-38f5695c3b5a.png)

![img](..\assets\ad112ff6-cafd-4574-bc7e-b5d4d04f8806.png)

那 CompletionService 是怎么做到获取最先执行完的任务结果的呢?

> 就是一个将异步任务的生产和任务完成结果的消费解耦的过程，利用消息队列，哪个任务执行的完，就直接将执行结果放到队列中，这样消费者拿到的结果自然就是最早完成的了

ExecutorCompletionService还有个构造方法，可以传入一个BlockingQueue<Future<V>> 所有的任务的Future结果都会加入到这个阻塞Queue中

<span style="background-image: linear-gradient(to right, rgb(255, 0, 0),rgb(0, 255, 0));color: white;padding: 0 4px;border-radius: 4px;letter-spacing:1px;">不要所有业务共用一个线程池</span> 隔离性能避免几个特别耗时的任务拖垮整个应用的风险

**主要场景**：

+ 上面的例子，有一组针对某个问题的solvers
+ 从多个服务下载，如果有一个下载成功，则终止其他下载过程

#### ScheduledExecutorService

ScheduledExecutorService可以说是Timer的替代类，因为Timer不支持多线程，任务是串行的，而且也不捕获异常，假设某个任务异常了，整个Timer就无法运行了

```java
ScheduledExecutorService service = Executors.newScheduledThreadPool(10);
service.scheduleAtFixedRate(new Runnable() {
	@Override
     public void run() {
     }
}, 1, 10, TimeUnit.SECONDS);//首次延迟1秒，之后每10秒执行一次（可能第一个任务10秒还没结束，不管它，到10秒再次执行）
```

#### ThreadPoolExecutor

```java
/**
  RUNNING   线程池正在运行，可以正常的接收新任务，同时执行任务队列中缓存的任务
  SHUTDOWN  线程池处于关闭状态，暂停接受新任务，但是会继续执行缓存队列中的任务
  STOP      线程池处于停止状态，暂停接受新任务，也不会去执行缓存队列中的旧任务
  TERMINATED线程池处于终止状态，不接受新任务，也不执行缓存队列中的旧任务，同时关闭所有执行线程
*/
volatile int runState; 
```

**拒绝策略**

实现`RejectedExecutionHandler`接口，JDK提供了四个实现类

| 策略                |                                                              |
| ------------------- | ------------------------------------------------------------ |
| CallerRunsPolicy    | 用调用者的线程，执行新的任务，如果任务执行是有严格次序的，请不要使用 |
| AbortPolicy         | 丢弃新任务，抛出RejectedExecutionException                   |
| DiscardPolicy       | 丢弃新任务，不通知调用者，多用于处理网络报文                 |
| DiscardOldestPolicy | 丢弃最旧的任务，处理网络报文时，可以使用此任务，因为报文处理是有时效的，超过时效的，都必须丢弃 |

**任务执行线程**

Worker本身，是一个Runnable对象，不是Thread对象，它内部封装一个Thread对象，用此对象执行本身的run方法，而这个Thread对象则由ThreadPoolExecutor提供的ThreadFactory对象创建新的线程，源码如下：

```java
private volatile ThreadFactory threadFactory;
    
private Thread addThread(Runnable firstTask) {
    Worker w = new Worker(firstTask);
    Thread t = threadFactory.newThread(w);
    if (t != null) {
        w.thread = t;
        workers.add(w);
        int nt = ++poolSize;
        if (nt > largestPoolSize)
            largestPoolSize = nt;
    }
    return t;
}
```

将Worker和Thread分离的好处是，如果我们的业务代码，需要对于线程池中的线程，赋予优先级、线程名称、线程执行策略等其他控制时，可以实现自己的ThreadFactory进行扩展，无需继承或改写ThreadPoolExecutor

**ThreadFactory**

```java
/**
 * Using thread factories removes hardwiring of calls to new Thread 避免硬引用
 * 
 * The simplest implementation of this interface is just:
 * class SimpleThreadFactory implements ThreadFactory {
 *   public Thread newThread(Runnable r) {
 *     return new Thread(r);
 *   }
 * }
 tomcat中的实现org.apache.tomcat.util.threads.TaskThreadFactory 写的非常好，可借鉴
 jdk实现：java.util.concurrent.Executors.DefaultThreadFactory
 */
public interface ThreadFactory {
    Thread newThread(Runnable r);
}
```

**扩展线程池**

通过继承ThreadPoolExecutor，并重写beforeExecute，afterExecute和terminated方法，我们可以在任务执行前，执行后和线程池关闭前干一些事情。如监控任务的平均执行时间，最大执行时间和最小执行时间等

**线程池的停止和关闭**

- shutdown()：只是将线程池的状态设成SHUTDOWN，然后中断所有没有正在执行任务的线程，等workQueue中所有任务都执行完后才终止，此时不会接受新的任务
- shutdownNow() 先将线程池的状态设置成STOP，遍历线程池中的工作线程，然后逐个调用线程的interrupt方法来中断线程（所以无法响应中断的任务可能永远无法终止）。并且清空workQueue，并返回等待执行任务的列表

**任务执行线程初始化**

worker线程，默认是在接受到新任务，或者workQueue满时，才创建，在实际使用中，可能需要在完成ThreadPoolExecutor对象实例化后，立即初始化所有任务执行线程，提高系统高峰期的吞吐量。上述需求，可以通过如下两个API实现：

```java
//初始化一个核心线程，如果核心线程数已达到上限，不创建新线程
public boolean prestartCoreThread() {
	return addIfUnderCorePoolSize(null);
}
//初始化所有核心线程
public int prestartAllCoreThreads() {
    int n = 0;
    while (addIfUnderCorePoolSize(null))
        ++n;
    return n;
}
```

**总结**
ThreadPoolExecutor是一个相当优秀的线程池，而且可以通过注入RejectHandler、ThreadFactory，BlockingQueue等方式，对其进行功能扩展，或者override beforeExecute和afterExecute方法进行运行task前后编织切面，比如实现暂停执行，恢复执行feature等

### Exchanger

> 可用于两个线程之间交换信息。可简单地将Exchanger对象理解为一个包含两个格子的容器，通过exchanger方法可以向两个格子中填充信息。当两个格子中的均被填充时，该对象会自动将两个格子的信息交换，然后返回给线程，从而实现两个线程的信息交换。

当超过两个线程调用同一个exchanger对象时，得到的结果是随机的，exchanger对象仅关心其包含的两个“格子”是否已被填充数据，当两个格子都填充数据完成时，该对象就认为线程之间已经配对成功，然后开始执行数据交换操作。

场景：线程交换信息

```java
Exchanger<Object> ex = new Exchanger();

new Thread(new Runnable {
	public void run() {
		System.out.println(exchanger.exchange("111111111"));
	}
}).start();
new Thread(Runnable {
	public void run() {
		System.out.println(exchanger.exchange("2222222222"));
	}
}).start();
```

### CountDownLatch

**倒数计时器**，使用场景如**多线程下载**，等待所有下载线程执行完，主线程才进行合并

![image](.\assets\CountdownLatch示意图.png)

latch.countDown()到了0，再调用latch.countDown()，不会报错；但是latch.getCount() 始终返回0

```java
final CountDownLatch begin = new CountDownLatch(1);
final CountDownLatch mainThreadLatch = new CountDownLatch(10);
final ExecutorService pool = Executors.newFixedThreadPool(10);
for (int i = 0; i < 10; i++) {
	pool.submit(new Runnable(){
		public void run() {
			begin.await();   //block当前线程，直到计数器为0或响应interrupt中断
			Thread.sleep((long) (Math.random() * 10000));
			mainThreadLatch.countDown(); //计数减1
		}
	});
}
mainThreadLatch.await(); //阻塞主线程，等计数器为0的时候,才被唤醒
pool.shutdown()
```

### CyclicBarrier

you can reuse CyclicBarrier once barrier is broken, but you can not reuse ContDownLatch

![image](.\assets\image-20200113103813985.png)

~~~java
CyclicBarrier barrier = new CyclicBarrier(3,new Runnable(){
	public void run(){
		//当计数器一次计数完成后执行的动作
	}
});
class Person implements Runnable {
	private int time;
	private CyclicBarrier barrier;
	private String name;
	public Person(CyclicBarrier barrier， String name， int time) {
		this.time = time;
		this.name = name;
		this.barrier = barrier;
	}
	public void run() {
		Thread.sleep(time * 1000);
		System.out.println(name + "到达电影院");
		barrier.await(); //计数减一，线程阻塞，直到所有线程都到达屏障点，CyclicBarrier会唤醒全部正在await()中等待的线程，此时屏障被打破，重新初始化成3
		//barrier.await(5,TimeUnit.MINUTES) 如果等待了5分钟其余线程还没有到达屏障，则认为屏障被broken，所有正在await()的线程被中断，会收到 BrokenBarrierException
		System.out.println("所有人都到电影院了，可以看了");
	}
}
ExecutorService pool = Executors.newFixedThreadPool(3);
pool.submit(new Person(barrier， "爸爸"， 20));
pool.submit(new Person(barrier， "妈妈"， 60));
pool.submit(new Person(barrier， "我"， 35));
~~~

### Semaphore

实现了标准的Dijkstra算法的计数信号量。允许多个线程同时进入临界区。可以认为信号量是一个共享锁，但是共享的额度是有限制的，额度用完了，其他没有拿到额度的线程要阻塞在临界区外。当额度为1时，就相等于互斥锁

方法：

+ acquire(int permits)：会阻塞线程，直到返回一个许可权
+ release()：将一个许可还给信号量

场景：排队上厕所，厕所只有两坑，来了10个人排队

~~~java
public class MySemaphore{
	Semaphore wcPosition = new Semaphore(2);
    
	public void test(){
		if(wcPosition.availablePermits()>0) 
			echo("有空位");
		else 
			echo("没空位，排队");
        
 		position.acquire(); //如果没有许可，会block
     	System.out.println("获得坑位");
     	Thread.sleep((int)(Math.random()*1000));
     	System.out.println("拉完屎");
     	position.release();
	}
}
~~~

# JDK8

## Stream

### 停止Steam循环

`stream.anyMatch(predicate)`

### 链式调用

```java
List<Integer> list = Lists.newArrayList(1,2,3,4,5,6,7,8,9,10)
list.stream()
    .filter(a->a!=null)
    .distinct()
    .mapToInt(a->a*2)
    .skip(2)
    .limit(4)
    .peek(System.out:println)
    .sum()
hashmap.forEach((key, value) -> {})
```

### 最小值

~~~java
String[] array = "1,2,3,4,5".split(",")
IntStream.of(array).min().getAsInt()
~~~

### 转大写

```java
List<String> list = Arrays.asList("a", "b", "c", "d");
list.stream()
    .map(String::toUpperCase)
    .collect(Collectors.toList())
    .forEach(System.out::println)
```

### list转map

```java
List<Account> accounts = new ArrayList();
Stream<TbBbs> stream = accounts.stream()
Map<Long, String> m = stream.collect(Collectors.toMap(Account::getId, Account::getUsername))
Map<Long, Account> m = stream.collect(Collectors.toMap(Account::getId, Function.identity()))
```

- toMap有个重载方法，可以传入一个合并的函数来解决key冲突问题：
  Map<Long, Account> m = stream.collect(Collectors.toMap(Account::getId, Function.identity(), (key1, key2) -> key2))
- toMap还有另一个重载方法，可以指定一个Map的具体实现
  stream.collect(Collectors.toMap(Account::getId, Function.identity(), (key1, key2) -> key2, LinkedHashMap::new))

### 对象集合转String集合

~~~java
List<String> fieldList = Stream.of(clazz.getDeclaredFields())
                .map(Field::getName)
                .collect(Collectors.toList());
~~~



### 数组转List

```java
Stream<String> stream = Stream.of("1,2,3".split(","))
List<String> l = stream.collect(Collectors.toList())
List<String> l = Arrays.asList("1,2,3".split(","))
```

### List转数组

```java
String[] arr = list.stream().toArray(String[]::new)
String[] arr = list.toArray(new String[list.size()])
```

## Optional

**什么是Optional类**

在日常工作中，各种变量、集合使用的过程中，往往要进行空判断，不然会出现NPE(Null Pointer Exception)

Optional类实际上是个容器：它可以保存类型T的值，或者仅仅保存null。Optional 类的引入很好的解决空指针异常。Optional提供很多有用的方法，这样我们就**不用显式进行null判断**

~~~java
// 参数不能是null  
Optional<Integer> o1 = Optional.of(1);  
// 参数可以是null  
Optional<Integer> o2 = Optional.ofNullable(null);  
// 参数可以是非null  
Optional<Integer> o3 = Optional.ofNullable(2);
//判断值是否存在
o1.isPresent()
//如果optional对象保存的值不是null，则返回原来的值，否则返回value
o1.orElse(value)
//值不存在则抛出异常，存在则什么不做
o1.orElseThrow(()->{new IllegalStateException();});  
~~~

判断是否为空

~~~java
Optional.ofNullable(name).orElse("")  //name为空的时候返回""
~~~

遍历list对象

~~~java
Optional.ofNullable(list).orElse(new ArrayList<>()).forEach(tmp->{});
~~~

获取对象的属性

~~~java
public String getName(User user) throws IllegalArgumentException{
	return Optional.ofNullable(user)
          .map(u->u.getRole())
          .map(r->r.getId())
          .orElseThrow(()->new IllegalArgumentException());
}

public User getUser(User user) {
	return Optional.ofNullable(user)
        .orElseGet(()-> {
			User user1 = new User();
			user1.setName("zhangsan");
			return user1;
		});
}
// orElseGet和orElse相似，都可以自定义类型返回，前者支持lamaba表达式
~~~



# 排序

### List排序

~~~java
List<Person> list = new ArrayList<>();
Collections.sort(list, new Comparator<Person>() {
	@Override
	public int compare(Person arg0, Person arg1) {
		return arg0.getId().compareTo(arg1.getId()); //升序
	}
});

//或实现接口
public class Person implements Comparable<Person>{
	@Override
    public int compareTo(Person arg0) {
        return this.getOrder().compareTo(arg0.getOrder()); //升序
    }
}
Collections.sort(list);
~~~

java8新接口

~~~java
//reversed 逆序
Collections.sort(all, Comparator.comparingInt(TbCmsInfo::getPraise).reversed());
~~~



### Map排序

~~~java
//按照key排序，用TreeMap即可

//按value排序
Map<String, Integer> map = new HashMap<String, Integer>();
List<Map.Entry<String, Integer>> l1 = new ArrayList<Map.Entry<String, Integer>>(map.entrySet());

Collections.sort(l1, new Comparator<Map.Entry<String, Integer>>() {   
	@Override
	public int compare(Map.Entry<String, Integer> o1, Map.Entry<String, Integer> o2) {   
     	return (o2.getValue() - o1.getValue()); //降序
		//return (o1.getKey()).toString().compareTo(o2.getKey());//比较key
    }
});
~~~

# 正则

~~~java
Pattern pattern = Pattern.compile("\\s+src=[^>]*", Pattern.CASE_INSENSITIVE);
Matcher matcher = pattern.matcher("<a src=\"https://www.baidu.com/bd_logo1.png\">");
if (matcher.find()){
	System.out.println(matcher.group());
}

//emoji表情 多个标志
Pattern emoji = Pattern.compile ("[\ud83c\udc00-\ud83c\udfff]|[\ud83d\udc00-\ud83d\udfff]|[\u2600-\u27ff]",Pattern.UNICODE_CASE | Pattern.CASE_INSENSITIVE );
~~~

> **Pattern Flag**

+ CASE_INSENSITIVE 大小写无关
+ UNICODE_CASE     当与CASE_INSENSITIVE结合时，使用Unicode字母匹配
+ MULTILINE             ^和$匹配一行的开始和结尾，而不是整个输入
+ UNIX_LINES           当在多行模式下匹配^和$时，只将'\n'看作行终止符
+ DOTALL                  当使用此标志时，.符号匹配包括行终止符在内的所有字符

~~~
[xyz]  x或y或z
[^xyz] 不匹配x，y，z
^      匹配行头
$      匹配行尾
.      匹配除 "\n" 外的所有字符
\d     0到9的数字  [0-9]
\D    非数字       [^0-9]
\s    空白字符     [\t\n\r\f\v\x0B]
\S    非空白字符    [^\t\n\r\f\v\x0B]
\w    字母，数字或下划线  [a-zA-Z0-9]
\W    [^a-zA-Z0-9]
a*    匹配a 0或多次
a+     匹配a 1或多次
a?     匹配a 0或1次
a{5}   匹配5次
a{5,}  匹配5+次
a{5,8} 匹配5到8次
中文   [\u4e00-\u9fa5]
~~~

## 

# I/O

## NAT内网穿透

内网机器如果不主动向外发起socket连接，外网机器是没法访问内网的IP的

参考`netapp`软件原理，

1. A是外网机器，B是内网机器，需要一个server做agent
2. 当A想连B时，A对agent说：替我告诉B，我想与它连接
3. agent对B说：你主动给A发一个请求。此时A和B就建立起tcp连接了（以后就不用agent帮忙了）

## 0.0.0.0

`0.0.0.0` 表示本机上所有的IPv4地址

如果一个机器有两个IP地址：192.168.1.1和10.1.2.1，配置一个服务监听0.0.0.0地址，那么用这两个ip都能连接上服务，即监听了本机所有的IP地址

若一个服务监听`127.0.0.1`，远程ip是无法连接这个服务的

## TIME_WAIT

+ 如果client主动关闭socket，在client的机器上使用netstat -an会看到很多TIME_WAIT的TCP链接
+ 如果server主动关闭，这种情况就出现在server的机器上
+ TCP协议中TIME_WAIT状态主要是为了保证数据的完整传输。一般会维持几分钟，TIME_WAIT 并不会占用很大资源，不用担心

## socket与最大打开文件数

+ 一台机器的端口个数为65535（0~1024系统保留，1024~65535是给用户用的）
+ Linux中，一个socket就是一个文件描述符，能打开多少个文件，就能创建多少socket（ulimit -n，默认一个进程最多打开1024个文件）

**结论**：在一个端口上（如8080），可以建立百万连接（netty）。一个机器最多能用65535个端口

# 集合

## LinkedHashMap

特性：**有序**（线程不安全）

迭代器遍历的方向是从链表的head节点到链表的tail节点（这个链表的顺序会变化的哦~）

~~~java
public class LinkedHashMap<K,V> extends HashMap<K,V> implements Map<K,V>{
	//The head (eldest) of the doubly linked list
	transient LinkedHashMap.Entry<K,V> head;
	//The tail (youngest) of the doubly linked list
	transient LinkedHashMap.Entry<K,V> tail;

    //true：访问排序，即最近最少使用的顺序，当调用get(key)或put(key)的时候，会把key元素移到链表的尾部。false：插入排序，就是put的时候是什么顺序，取出来的时候就是什么顺序
	final boolean accessOrder;
}
~~~

![image-20200115211513137](.\assets\image-20200115211513137.png)



## LRUHashMap

LRUMap对于构建缓存或者连接池之类的技术经常用到，common-collections框架给了现成的实现

## TreeMap

底层数据结构：**红黑树**（即一颗**自平衡的排序二叉树**），key是有序的

二叉树基本性质：树的任何节点的值大于它的左子节点，小于它的右子节点。但是在生成二叉树的过程是非常容易失衡的，最坏的情况就是一边倒（只有右/左子树），这样导致二叉树的检索效率大大降低（复杂度为O(n)），所以为了维持二叉树的平衡，大牛们提出了各种算法，如：AVL，SBT，伸展树，红黑树。

平衡二叉树必须具备：它的左右两个子树的高度差的绝对值不超过1，左右两个子树都是一棵平衡二叉树

红黑树顾名思义就是节点是红色或者黑色的平衡二叉树，它通过颜色的约束来维持二叉树的平衡。如下规则

1. 每个节点都只能是红色或黑色
2. 根节点是黑色
3. 每个叶节点（NIL节点，空节点）是黑色的
4. 如果一个结点是红的，则它两个子节点都是黑的
5. 从任一节点到其每个叶子的所有路径都包含相同数目的黑色节点

检索效率O(logN)

![image](.\assets\image-20200115212322445.png)

三大基本操作：左旋、右旋、着色

![image-20200115212447934](.\assets\image-20200115212447934.png)

![image-20200115212517880](.\assets\image-20200115212517880.png)

增加节点

+ 插入新节点总是红色
+ 如果插入节点的父节点是黑色, 能维持性质 
+ 如果插入节点的父节点是红色, 破坏了性质. 故插入算法就是通过重新着色或旋转, 来维持性质4 [具体如何着色和旋转见](https://blog.csdn.net/eson_15/article/details/51144079?utm_source=blogxgwz0)

ArrayList：线程不安全，Vector：线程安全

HashMap：线程不安全：CocurrentHashMap：线程安全

**集合的安全删除**

在集合的Iterator上调用 `remove()` 安全地删除对象，避免 ConcurrentModifiedException

# Java基础

### 关键字

#### final

+ `final`修饰的类不能被继承
+ `final`修饰的方法不能被子类重写（编译器在遇到调用final方法时会转入内嵌机制，提高执行效率）
+ `final`修饰的变量是常量，只能被赋值一次

#### static

+ static是属于类的，不属于对象，所以static修饰的方法默认是final的
+ 既然是final的，所以static方法不能被子类重写
+ 子类也不能把父类不是静态的方法重写成静态的方法

#### transient

+ 修饰的变量不会被序列化，用于修饰密码等字段

#### finally

~~~java
try{
	if(flag){
        System.exit(0);//终止当前JVM进程
    }else{
        return 123; //finally 在 return之前执行。finally不能试图改变return的返回值
    }
}finally{ 
	//不被执行的唯一情况是执行了System.exit，因为进程都没了，咋执行
}
~~~

### 泛型

~~~java
//<T>修饰符表示泛型方法
public <T> T method(String s, Class<T> clazz, Object... args) {
	Assert.hasText(s, "不能为空");
  Assert.isTrue(clazz != null, "clazz 不能为空");
}
public <T> List<T> method(Class<T> clazz){
}
public <T> T set(Object key, T t) {
	return t;
}
~~~



### Inner Class

> Why use Inner Class？内部类可以实现《多重继承》，有成员内部类，局部内部类，匿名内部类和静态内部类

成员内部类和局部内部类

~~~java
public class Test{
    private String name = "out-name";
    private String id = "111";
    //成员内部类
    class Entry{
        private String name = "inner-name";
        private void print(){
            System.out.print(id);            //内部类能直接访问外部类的成员
            System.out.print(name);          //inner-name
            System.out.print(Test.this.name);//out-name
        }
    }
    
    private AI run(){
        //局部内部类，作用域仅限于方法内
        class AI{
            public void go(){}
        }
        //记录log等业务
        return new AT()
    }
}
~~~

静态内部类

~~~java
public class Test{
    //静态内部类是不需要依赖于外部类
    static class Entry{
        private String name = "inner-name";
    }
}
//call
Test.Entry te = new Test.Entry()
~~~

匿名内部类一般用于回调，当go方法执行完，局部变量`name`的生命周期就结束了，而内部类对象的生命周期可能还没有结束，那么在内部类中访问name就不可能了，所以将name改为final，改变其生命周期

~~~java
public void go(final String name){
    new Thread(new Runnable(){
    	//当前类为匿名内部类
    	@Override
    	public void run() {
        	System.out.print(name)
		} 
	}).start()
}
~~~

### 枚举

~~~java
public enum Color {
	RED, GREEN, BLANK, YELLOW
}
switch(color){
    case RED:
    case GREEN:
}
~~~

另一种写法

~~~java
public enum Color {
	RED("红色", 1), GREEN("绿色", 2), BLANK("白色", 3);

	private String name;
	private int index; 
	private Color(String name, int index) {
		this.name = name;
		this.index = index;
	}
	public static String getName(int index) {
		for (Color c : Color.values()) {
			if (c.getIndex() == index)
				return c.name;
		}
		return null;
	}
}
~~~

### 注解

~~~java
@Target({ElementType.TYPE,ElementType.METHOD}) 
@Retention(RetentionPolicy.RUNTIME) 
@Documented
@Inherited
public @interface Skip {
	String value() default "965";
}
~~~

解析注解

~~~java
Class clazz = AnnotationParsing.class.getClassLoader().loadClass("com.iqiyi.TestAction");
Class clazz = obj.getClass();

Method method = clazz.getMethods()[0];
method.isAnnotationPresent(Skip.class); //此方法上是否有Skip注解
method.getDeclaredAnnotations(); //获得此方法上所有的注解

Field field = clazz.getDeclaredField("id");
Annotation skipAnno = field.getAnnotation(Skip.class);
~~~



### 对象克隆

Object默认是“浅拷贝”，即不会克隆对象中的对象，为了实现对象的深烤贝，应该重写Object类的clone()

~~~java
class Person implements Cloneable {
	private String name;

	@Override
	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}
}
~~~

### hashCode

判断object是否被修改过（金蝶面试）：判断hashCode是否变化

2个对象equals相等，那么它们的hashcode一定相同

2个对象hashcode相等，他们equals不一定是相等的

覆盖equals()时必须同时覆盖hashCode()，否则会出问题

### Exception机制

Error：没办法处理

Exception分为check异常和uncheck异常，比如IOException异常就是可以check的，NullPointerException是uncheck的，没法检查这类异常，一旦出现就会程序终止，类似的还有ArrayIndexOutOfBoundsException

### String

~~~java
//将字符串添加到驻留池中，并返回这个已驻留字符串的引用
public native String intern();	

String str2 = new String(“张三”);
str2 = str2.intern();


//JVM会新建一个StringBuilder对象，用append完成+的逻辑
//缺点：+产生了大量的StringBuilder临时对象
"aaa"+new String("bbb");
~~~

StringBuffer：  线程安全，可变
StringBuilder：非线程安全，可变
String：             不可变
性能：StringBuilder > StringBuffer > String

### Reference

#### 强引用

只要对象有引用，就不会被GC

#### 弱引用

java.lang.ref.WeakReference

**场景**：写一个Java程序，观察某对象什么时候会被gc，你必须要用一个reference记住此对象，以便随时观察，但是却因此造成此对象的reference数目一直无法为零，使得对象无法被清除，能随时取得某对象的信息，但又不想影响此对象的gc，此时就应该用弱引用来记住此对象

~~~java
WeakReference wr= new WeakReference(obj);
obj=null;
System.gc();
if(wr.get()==null){
	System.out.println("obj已被gc");
}else{
	System.out.println("obj尚未被GC," + wr.get().toString());
}
~~~

在设计Optimizer或Debugger这类的程序时常会用到，因为这类程序需要取得某对象的信息，但是不可以影响此对象的垃圾收集

#### 软引用

java.lang.ref.SoftReference
SoftReference与WeakReference的特性基本一致， 最大的区别在于SoftReference会尽可能长的保留引用直到JVM内存不足时才会被gc，被SoftReference指到的对象，即使没有任何强引用，也不会被清除

**gc时机**：内存不足时且没有其他强引用

**目的**：设计object-cache。**使用软引用能防止OOM**

例子：Android手机开发多用做imageCache

~~~java
Map<String, SoftReference<Bitmap>> imageCache = new new HashMap<String, SoftReference<Bitmap>>();
Bitmap bitmap = BitmapFactory.decodeStream(InputStream);//强引用的Bitmap对象
SoftReference<Bitmap> soft = new SoftReference<Bitmap>(bitmap);//软引用的Bitmap对象，如果内存不足，这些软引用对象会被gc
imageCache.put("1",soft);
~~~

#### 虚引用

java.lang.ref.PhantomReference，永远访问不到虚引用指向的对象，又叫幽灵引用

# JVM

## Class Loader

**作用**：Class Loader只负责加载.class文件到内存，格式正确就能加载，至于能不能运行，由执行引擎负责

加载模式：双亲委托

使用 `jclasslib bytecode viewer` 可以查看class文件内容

~~~java
ClassFile { //Class文件的规范
      u4 magic;
      u2 minor_version;
      u2 major_version;
      u2 constant_pool_count;
      cp_info constant_pool[constant_pool_count-1];
      u2 access_flags;
      u2 this_class;
      u2 super_class;
      u2 interfaces_count;
      u2 interfaces[interfaces_count];
      u2 fields_count;
      field_info fields[fields_count];
      u2 methods_count;
      method_info methods[methods_count];
      u2 attributes_count;
      attribute_info attributes[attributes_count];
}
~~~

## PC Register

每一个线程都有一个计数器，指向方法区中的方法（即程序执行到字节码的行号)。计数器解决了多线程切换后能返回现场

## Stack

**生命周期**：与线程生命周期相同，不存在gc，因为线程结束，栈内存就释放了

**栈中都存的什么数据**

+ 局部变量：如输入输出参数、基本数据类型，对象引用，方法返回地址等
+ 栈操作（记录出栈、入栈的操作）
+ 栈帧数据（包括类文件、方法等等）
+ 方法索引表：记录的是每个method的地址

**Java传参是传值调用**

基本类型和引用的处理是一样的，都是传值，所以，如果是传引用的方法调用，就是“传引用值”的传值调用

**方法调用栈** (method invocation stack)
每一个方法的调用至完成，就意味着一个帧在VM栈中的入栈至出栈的过程
每一个方法被执行的时候，都会同时创建一个栈帧（Stack Frame）

## Heap

**堆是所有线程共享的**，里面存放的是对象实例和数组

堆内存划分为三部分：Permanent Generation、Young Generation、Old Generation，划分的目的是为了更好的分代gc

根据VM Spec的要求，Java堆可以处于物理上不连续的内存空间，它逻辑上是连续的即可

## 参数调优

工具

1. `jps -lm` 查看进程id，参数
2. `jstat -gc pid` 显示gc次数和时间
3. core dump：java线程信息转储 jstack -l pid > coredump.txt
4. heap dump：堆信息转储 jhat

JVM参数

|      | 参数                                             | 描述                                                    |
| ---- | ------------------------------------------------ | ------------------------------------------------------- |
| ☆    | -Xmn128m                                         | 新生代大小 = (Eden+S0+S1)，推荐占堆的1/3                |
| ☆    | -Xmx512m                                         | Java heap size的最大值                                  |
| ☆    | -Xms512m                                         | Java heap size的初始值                                  |
| ☆    | -XX:SurvivorRatio=8                              | Eden占新生代内存中的8份，0区和1区各占1份内存            |
|      | -XX:NewRatio=4                                   | 新生代内存:老年代内存=1:4    （老年代不含永久区）       |
|      | -XX:+UseParallelGC                               | 并行回收，用于：后台处理，科学计算，响应慢              |
|      | -XX:+UseParNewGC -XX:+UseConcMarkSweepGC         | 并发回收（多线程gc），用于：web服务器 响应快            |
|      | -XX:MetaspaceSize=512m -XX:MaxMetaspaceSize=768m | jdk1.8中持久代被移除了,取代的是元数据区(Metadata Space) |

## GC

### 再见PermGen，你好Metaspace

随着JDK8的到来，JVM不再有PermGen。但类的元数据信息（metadata）还在，只不过不再是存储在连续的堆空间上，而是移动到叫做“Metaspace”的本地内存（Native memory）中。

### 原理

- JVM通过有向图管理heap中的对象，检测对象是否被引用(引用计数和对象引用遍历)
- gc监控每一个对象的运行状态，包括对象的申请、引用、被引用、赋值等

`Runtime.getRuntime().gc()` 只建议jvm执行gc，不保证立即执行，因为gc线程优先级很低的哦~

### 引用计数

对象保存一个计数器，对象增加一个引用计数器就+1，删除一个引用计数器就-1。扫描计数器为0的对象，将它们回收。此算法最致命的是无法处理循环引用。

###  分代GC

分代的回收策略基于一个事实：不同的对象的生命周期不一样。不同生命周期的对象可采取不同的收集方式，以便提高回收效率。所以**不同代的gc算法也不一样**

**新生代GC**

1. 每次垃圾收集时都发现有大批对象死去，只有少量存活，用复制算法，只复制少量存活对象就可以完成收集
2. 尽可能快速的收集掉那些生命周期短的对象
3. -Xmn1024m

![1567647800470](../assets/jvm堆内存占比.png)

~~~
新生区
  ├─Eden space(伊甸区)         ... n
  └─Survivor space(幸存者区) 
      ├─0区(Survivor 0)       ... 1
      └─1区(Survivor 1)       ... 1
~~~

所有的类都是在`Eden`被new出来。当伊甸区满时，jvm对伊甸区gc，将存活对象移动到幸存0区。若0区也满了，对0区进行gc，然后存活对象移动到1区。那如果1区也满了呢？gc后再移动到0区，如此往复，在年轻代中经历了N次这样的gc后仍然存活的对象，就会被放到年老代中

**老年代GC**

1. 老年代中因为对象存活率高、没有额外空间供它分配，就必须使用《标记-清理》或《标记-整理》算法进行回收
2. 存放经过多次GC仍存活的对象(从新生区筛选出来的对象)
3. -Xmx1024m

**方法区GC**

1. 存字节码，接口的meta data，反射对象，static变量，**Constant Pool**（final常量）。运行期也可将新内容放入常量池`String.intern()`
2. VM Spec上描述方法区为堆的一个逻辑部分（the method area is logically part of the heap），**被所有线程共享**，**常驻内存**，不会被gc（如果类非常多，可能OOM）
3. **享元模式**

## 线上排错

### CPU利用率过高

+ us值过高
  1. 表示应用消耗了大部分CPU。主要是线程一直执行，无任何挂起动作，通常是这些线程执行无阻塞、循环、正则或数学计算等（解决：增加Thread.sleep(1000)）
  2. 另外一个可能是频繁的GC（参数调优，降低GC次数）
  3. 分析：top命令，shift+H显示线程信息，找到消耗CPU严重的线程ID，转成十六进制，然后jstack导出core dump文件，在里面查找pid分析

+ sy值过高
  1. 表示CPU一直在线程切换，主要是启动的线程过多，这些线程多数处于不断的阻塞（锁等待，I/O等待）和运行切换
  2. 线程间锁竞争激烈
  3. 分析同上

### OutOfMemory

+ 堆内存不足了
+ 永久区空间不足

## Direct Memory

NIO可以通过Native函数库直接分配OS内存，然后通过一个存储在Java堆里面的DirectByteBuffer对象作为这块内存的引用进行操作。避免在jvm和OS中来回复制数据

## 计算java对象的大小

**对象内存组成**

+ Object header ：非数组对象为8byte，数组为12byte（Object是所有对象的父类，所以一个对象的对象头就是指的这个，对象头为12，4bytes保存的是数组的length ）
+ primitive fields：
  + boolean byte 1byte
  + char short 2byte
  + int float 4byte
  + long double 8byte
+ reference fields：4 bytes 
+ padding ：为了内存对齐的填充字节。Hotspot规定对象占的字节数必须是8的整数倍 

~~~java
class A { 
    boolean val0; 
    long    val1; 
    int     val2; 
} 
size=8(header)+1(boolean)+8(long)+4(int)+3(padding)=24 
~~~

一个复杂的Java对象的空间占用是递归计算的，即对象类型本身->父类->祖先类->Object

## 字节码分析

~~~java
getstatic       #2; //获取常量池中索引为2所表示的类变量
invokevirtual   #3; //调用常量池索引为3的方法

public static void main(String[] args){
	int a=3;
  int b=5;
  a = b + 0*(b=a);
}

public static void main(java.lang.String[] args);
     0  iconst_3        --3入栈【3】
     1  istore_1 [a]    --栈顶弹出，存到局部变量区索引为1的地方（变量a）。【】a=3
     2  iconst_5        --5入栈【5】
     3  istore_2 [b]    --栈顶弹出，存到局部变量区索引为2的地方（变量b）。【】b=5
     4  iload_2 [b]     --局部变量区索引为2的值入栈【5】
     5  iconst_0        --0入栈【5，0】
     6  iload_1 [a]     --局部变量区索引为1的值入栈【5,0,3】
     7  dup             --栈顶复制一个字节内容入栈 【5,0,3,3】
     8  istore_2 [b]    --栈顶弹出，存到局部变量区索引为2的地方。【5,0,3】b=3
     9  imul            --弹出3,0，做乘法，为0入栈 【5,0】
    10  iadd            --弹出5,0，做加法 为5入栈【5】
    11  istore_1 [a]    --栈顶弹出，存到局部变量区索引为2的地方【】
    12  return
      Line numbers:
        [pc: 0, line: 3]
        [pc: 2, line: 4]
        [pc: 4, line: 5]
        [pc: 12, line: 6]
      Local variable table:
        [pc: 0, pc: 13] local: args index: 0 type: java.lang.String[]
        [pc: 2, pc: 13] local: a index: 1 type: int
        [pc: 4, pc: 13] local: b index: 2 type: int
}
~~~

# Servlet

## Forward

<span style='color:red'>服务器转发</span> ：服务器直接访问目标url，把那个url的响应读取过来，然后把这些内容再发给浏览器。浏览器根本不知道内容从哪里来的，所以它的地址栏还是原来的地址。

一般用于用户登陆的时候,根据角色转发到相应的模块

~~~java
@GetMapping("diagram/{processInstanceId}")
public void diagram(@PathVariable String processInstanceId, HttpServletRequest req, HttpServletResponse res) throws Exception {
    RequestDispatcher dispatcher = req.getRequestDispatcher("/runtime/process-instances/"+processInstanceId+"/diagram");
    req.setAttribute("key1","额外请求参数");
    dispatcher.forward(req,res);
}
~~~

## Redirect

<span style='color:green'>客户端重定向</span> ：server返回状态码（HttpServletResponse.SC_NOT_MODIFIED如304），告诉浏览器重新去请求转发的地址，所以地址栏显示的是新的URL。因为是重新发起的请求，所以不能共享数据

一般用于注销时返回主页面和跳转到其它的网站等

~~~java
response.sendRedirect("/main.html");
~~~



# 代码规范

## 打印异常日志

不要直接在catch里写`e.printStackTrace()`。因为这打印出来的错误**没有日期、等级**等。应该如下

~~~java
log.error("xxx has error, {}", e);
~~~

## Assert替换throw exception

~~~java
Assert.notNull(user, "使用断言");
~~~

## 字符串格式化

~~~java
MessageFormat.format("spring-schedule-{0}", id)
~~~

